@@ -1,6 +1,13 @@
 """Module containing the code for the problem solvers."""
 
 import random
+from types import FunctionType
+from constraint.domain import Domain
+from constraint.constraints import Constraint, FunctionConstraint, CompilableFunctionConstraint
+from collections.abc import Hashable
+
+# for parallel solver
+from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
 
 
 def getArcs(domains: dict, constraints: list[tuple]) -> dict:
@@ -65,6 +72,8 @@ def doArc8(arcs: dict, domains: dict, assignments: dict) -> bool:
 class Solver:
     """Abstract base class for solvers."""
 
+    requires_pickling = False
+
     def getSolution(self, domains: dict, constraints: list[tuple], vconstraints: dict):
         """Return one solution for the given problem.
 
@@ -332,7 +341,7 @@ class OptimizedBacktrackingSolver(Solver):
 
         raise RuntimeError("Can't happen")
 
-    def getSolutionsList(self, domains: dict, vconstraints: dict) -> list[dict]:  # noqa: D102
+    def getSolutionsList(self, domains: dict[Hashable, Domain], vconstraints: dict[Hashable, list[tuple[Constraint, Hashable]]]) -> list[dict[Hashable, any]]:  # noqa: D102, E501
         """Optimized all-solutions finder that skips forwardchecking and returns the solutions in a list.
 
         Args:
@@ -343,6 +352,7 @@ class OptimizedBacktrackingSolver(Solver):
             the list of solutions as a dictionary.
         """
         # Does not do forwardcheck for simplicity
+
         assignments: dict = {}
         queue: list[tuple] = []
         solutions: list[dict] = list()
@@ -388,8 +398,6 @@ class OptimizedBacktrackingSolver(Solver):
             # Push state before looking for next variable.
             queue.append((variable, values))
 
-        raise RuntimeError("Can't happen")
-
     def getSolutions(self, domains: dict, constraints: list[tuple], vconstraints: dict):  # noqa: D102
         if self._forwardcheck:
             return list(self.getSolutionIter(domains, constraints, vconstraints))
@@ -417,7 +425,6 @@ class OptimizedBacktrackingSolver(Solver):
         lst.sort()
         return [c for _, _, c in lst]
 
-
 class RecursiveBacktrackingSolver(Solver):
     """Recursive problem solver with backtracking capabilities.
 
@@ -595,3 +602,187 @@ class MinConflictsSolver(Solver):
             if not conflicted:
                 return assignments
         return None
+
+
+### Helper functions for parallel solver
+
+def is_valid(assignment: dict[Hashable, any], constraints_lookup: list[tuple[Constraint, Hashable]], domains: dict[Hashable, Domain]) -> bool:      # noqa E501
+    """Check if all constraints are satisfied given the current assignment."""
+    return all(
+        constraint(vars_involved, domains, assignment, None)
+        for constraint, vars_involved in constraints_lookup
+        if all(v in assignment for v in vars_involved)
+    )
+
+def compile_to_function(constraint: CompilableFunctionConstraint) -> FunctionConstraint:
+    """Compile a CompilableFunctionConstraint to a function, wrapped by a FunctionConstraint."""
+    func_string = constraint._func
+    code_object = compile(func_string, "<string>", "exec")
+    func = FunctionType(code_object.co_consts[0], globals())
+    return FunctionConstraint(func)
+    
+def sequential_recursive_backtrack(assignment: dict[Hashable, any], unassigned_vars: list[Hashable], domains: dict[Hashable, Domain], constraint_lookup: dict[Hashable, list[tuple[Constraint, Hashable]]]) -> list[dict[Hashable, any]]:     # noqa E501
+    """Sequential recursive backtracking function for subproblems."""
+    if not unassigned_vars:
+        return [assignment.copy()]
+
+    var = unassigned_vars[-1]
+    remaining_vars = unassigned_vars[:-1]
+
+    solutions: list[dict[Hashable, any]] = []
+    for value in domains[var]:
+        assignment[var] = value
+        if is_valid(assignment, constraint_lookup[var], domains):
+            solutions.extend(sequential_recursive_backtrack(assignment, remaining_vars, domains, constraint_lookup))
+        del assignment[var]
+    return solutions
+
+def sequential_optimized_backtrack(assignment: dict[Hashable, any], unassigned_vars: list[Hashable], domains: dict[Hashable, Domain], constraint_lookup: dict[Hashable, list[tuple[Constraint, Hashable]]]) -> list[dict[Hashable, any]]:     # noqa E501
+    """Sequential optimized backtracking (as in OptimizedBacktrackingSolver) function for subproblems."""
+    # Does not do forwardcheck for simplicity
+
+    assignments = assignment
+    sorted_variables = unassigned_vars
+    queue: list[tuple] = []
+    solutions: list[dict] = list()
+
+    while True:
+        # Mix the Degree and Minimum Remaing Values (MRV) heuristics
+        for variable in sorted_variables:
+            if variable not in assignments:
+                # Found unassigned variable
+                values = domains[variable][:]
+                break
+        else:
+            # No unassigned variables. We've got a solution. Go back
+            # to last variable, if there's one.
+            solutions.append(assignments.copy())
+            if not queue:
+                return solutions
+            variable, values = queue.pop()
+
+        while True:
+            # We have a variable. Do we have any values left?
+            if not values:
+                # No. Go back to last variable, if there's one.
+                del assignments[variable]
+                while queue:
+                    variable, values = queue.pop()
+                    if values:
+                        break
+                    del assignments[variable]
+                else:
+                    return solutions
+
+            # Got a value. Check it.
+            assignments[variable] = values.pop()
+            for constraint, variables in constraint_lookup[variable]:
+                if not constraint(variables, domains, assignments, None):
+                    # Value is not good.
+                    break
+            else:
+                break
+
+        # Push state before looking for next variable.
+        queue.append((variable, values))
+
+
+def parallel_worker(args: tuple[bool, dict[Hashable, Domain], dict[Hashable, list[tuple[Constraint, Hashable]]], Hashable, any, list[Hashable]]) -> list[dict[Hashable, any]]:    # noqa E501
+    """Worker function for parallel execution on first variable."""
+    process_mode, domains, constraint_lookup, first_var, first_value, remaining_vars = args
+    local_assignment = {first_var: first_value}
+
+    if process_mode:
+        # if there are any CompilableFunctionConstraint, they must be compiled locally first
+        for var, constraints in constraint_lookup.items():
+            constraint_lookup[var] = [tuple([compile_to_function(constraint) if isinstance(constraint, CompilableFunctionConstraint) else constraint, vals]) for constraint, vals in constraints]        # noqa E501
+
+    # continue solving sequentially on this process
+    if is_valid(local_assignment, constraint_lookup[first_var], domains):
+        return sequential_optimized_backtrack(local_assignment, remaining_vars, domains, constraint_lookup)
+    return []
+
+class ParallelSolver(Solver):
+    """Problem solver that executes all-solution solve in parallel (ProcessPool or ThreadPool mode).
+
+    Sorts the domains on size, creating jobs for each value in the domain with the most variables.
+    Each leaf job is solved locally with either optimized backtracking or recursion.
+    Whether this is actually faster than non-parallel solving depends on your problem, and hardware and software environment. 
+
+    Uses ThreadPool by default. Instantiate with process_mode=True to use ProcessPool. 
+    In ProcessPool mode, the jobs do not share memory.
+    In ProcessPool mode, precompiled FunctionConstraints are not allowed due to pickling, use string constraints instead.
+
+    Examples:
+        >>> result = [[('a', 1), ('b', 2)],
+        ...           [('a', 1), ('b', 3)],
+        ...           [('a', 2), ('b', 3)]]
+
+        >>> problem = Problem(ParallelSolver())
+        >>> problem.addVariables(["a", "b"], [1, 2, 3])
+        >>> problem.addConstraint("b > a", ["a", "b"])
+
+        >>> for solution in problem.getSolutions():
+        ...     sorted(solution.items()) in result
+        True
+        True
+        True
+
+        >>> problem.getSolution()
+        Traceback (most recent call last):
+        ...
+        NotImplementedError: ParallelSolver only provides all solutions
+
+        >>> problem.getSolutionIter()
+        Traceback (most recent call last):
+        ...
+        NotImplementedError: ParallelSolver doesn't provide iteration
+    """     # noqa E501
+
+    def __init__(self, process_mode=False):
+        """Initialization method. Set `process_mode` to True for using ProcessPool, otherwise uses ThreadPool."""
+        super().__init__()
+        self._process_mode = process_mode
+        self.requires_pickling = process_mode
+
+    def getSolution(self, domains: dict, constraints: list[tuple], vconstraints: dict):
+        """Return one solution for the given problem.
+
+        Args:
+            domains (dict): Dictionary mapping variables to their domains
+            constraints (list): List of pairs of (constraint, variables)
+            vconstraints (dict): Dictionary mapping variables to a list
+                of constraints affecting the given variables.
+        """
+        msg = f"{self.__class__.__name__} only provides all solutions"
+        raise NotImplementedError(msg)
+
+    def getSolutionsList(self, domains: dict[Hashable, Domain], vconstraints: dict[Hashable, list[tuple[Constraint, Hashable]]]) -> list[dict[Hashable, any]]:  # noqa: D102, E501
+        """Parallelized all-solutions finder using ProcessPoolExecutor for work-stealing."""
+        # Precompute constraints lookup per variable
+        constraint_lookup: dict[Hashable, list[tuple[Constraint, Hashable]]] = {var: vconstraints.get(var, []) for var in domains}  # noqa: E501
+
+        # Sort variables by domain size (heuristic)
+        sorted_vars: list[Hashable] = sorted(domains.keys(), key=lambda v: len(domains[v]))
+
+        # Split parallel and sequential parts
+        first_var = sorted_vars[0]
+        remaining_vars = sorted_vars[1:]
+
+        # Create the parallel function arguments and solutions lists
+        args = ((self.requires_pickling, domains, constraint_lookup, first_var, val, remaining_vars.copy()) for val in domains[first_var])  # noqa: E501
+        solutions: list[dict[Hashable, any]] = []
+
+        # execute in parallel
+        parallel_pool = ProcessPoolExecutor if self._process_mode else ThreadPoolExecutor
+        with parallel_pool() as executor:
+            # results = map(parallel_worker, args)  # sequential
+            results = executor.map(parallel_worker, args, chunksize=1)   # parallel
+            for result in results:
+                solutions.extend(result)
+
+        return solutions
+    
+    def getSolutions(self, domains: dict, constraints: list[tuple], vconstraints: dict):  # noqa: D102
+        return self.getSolutionsList(domains, vconstraints)
+    
\ No newline at end of file
